import asyncio
import warnings
import logging
import random
from enum import IntEnum, Enum
from typing import Optional, TYPE_CHECKING

from feeluown.excs import ModelNotFound, ProviderIOError
from feeluown.utils import aio
from feeluown.utils.aio import run_fn, run_afn
from feeluown.utils.dispatch import Signal
from feeluown.utils.utils import DedupList
from feeluown.player import Metadata, MetadataFields
from feeluown.library import (
    MediaNotFound, SongModel, ModelType, ResourceNotFound, VideoModel,
)
from feeluown.media import Media
from feeluown.library import reverse

if TYPE_CHECKING:
    from feeluown.app import App

logger = logging.getLogger(__name__)


class PlaybackMode(IntEnum):
    """
    Playlist playback mode.

    .. versiondeprecated:: 3.8.12
        Please use PlaylistRepeatMode and PlaylistShuffleMode instead.
    """
    one_loop = 0  #: One Loop
    sequential = 1  #: Sequential
    loop = 2  #: Loop
    random = 3  #: Random


class PlaylistRepeatMode(Enum):
    none = 'none'
    one = 'one'
    all = 'all'


class PlaylistShuffleMode(Enum):
    """
    Windows and macOS has multiple shuffle modes.
    """
    off = 'off'
    songs = 'songs'


class PlaylistMode(IntEnum):
    """playlist mode

    **What is FM mode?**

    In FM mode, playlist's playback_mode is unchangeable, it will
    always be sequential. When playlist has no more song,
    the playlist hopes someone(we call it ``FM`` here) will:
    1. catch the ``eof_reached`` signal
    2. add news songs to playlist by using ``fm_add`` method

    **When will playlist exit FM mode?**

    If user manually play a song, playlist will exit FM mode, at the
    same time, playlist will:
    1. clear itself
    2. change to normal mode
    3. set current song to the song
    """
    normal = 0  #: Normal
    fm = 1  #: FM mode


class Playlist:
    def __init__(self, app: 'App', songs=None, playback_mode=PlaybackMode.loop,
                 audio_select_policy='hq<>'):
        """
        :param songs: list of :class:`feeluown.library.SongModel`
        :param playback_mode: :class:`feeluown.player.PlaybackMode`
        """
        self._app = app

        #: init playlist mode normal
        self._mode = PlaylistMode.normal

        #: playlist eof signal
        # playlist have no enough songs
        self.eof_reached = Signal()

        #: playlist mode changed signal
        self.mode_changed = Signal()

        #: store value for ``current_song`` property
        self._current_song = None
        self._current_song_mv = None

        #: songs whose url is invalid
        self._bad_songs = DedupList()

        #: store value for ``songs`` property
        self._songs = DedupList(songs or [])

        self.audio_select_policy = audio_select_policy

        #: store value for ``playback_mode`` property
        self._playback_mode = playback_mode

        #: playback mode changed signal
        self.playback_mode_changed = Signal()
        self.song_changed = Signal()
        """current song changed signal

        The player will play the song after it receive the signal,
        when song is None, the player will stop playback.
        """
        self.song_changed_v2 = Signal()
        """current song chagned signal, v2

        emit(song, media)
        """
        self.song_mv_changed = Signal()  # emit(song, mv)

        #: When watch mode is on, playlist try to play the mv/video of the song
        self.watch_mode = False

        self._t_scm = self._app.task_mgr.get_or_create('set-current-model')

        # .. versionadded:: 3.7.11
        #    The *songs_removed* and *songs_added* signal.
        self.songs_removed = Signal()  # (index, count)
        self.songs_added = Signal()  # (index, count)
        # .. versionadded:: 3.9.0
        #    The *play_model_handling* signal.
        self.play_model_handling = Signal()

        self._app.player.media_finished.connect(self._on_media_finished)
        self.song_changed.connect(self._on_song_changed)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        """set playlist mode"""
        if self._mode is not mode:
            if mode is PlaylistMode.fm:
                self.playback_mode = PlaybackMode.sequential
            # we should change _mode at the very end
            self._mode = mode
            self.mode_changed.emit(mode)
            logger.info('playlist mode changed to %s', mode)

    @property
    def repeat_mode(self):
        if self.playback_mode in (PlaybackMode.loop, PlaybackMode.random):
            return PlaylistRepeatMode.all
        elif self.playback_mode is PlaybackMode.one_loop:
            return PlaylistRepeatMode.one
        return PlaylistRepeatMode.none

    @repeat_mode.setter
    def repeat_mode(self, mode):
        if mode is PlaylistRepeatMode.all:
            if self.playback_mode not in (PlaybackMode.loop, PlaybackMode.random):
                self.playback_mode = PlaybackMode.loop
        elif mode is PlaylistRepeatMode.one:
            self.playback_mode = PlaybackMode.one_loop
        else:
            self.playback_mode = PlaybackMode.sequential

    @property
    def shuffle_mode(self):
        if self.playback_mode is PlaybackMode.random:
            return PlaylistShuffleMode.songs
        return PlaylistShuffleMode.off

    @shuffle_mode.setter
    def shuffle_mode(self, mode):
        if mode is PlaylistShuffleMode.songs:
            self.playback_mode = PlaybackMode.random
        else:
            self.playback_mode = PlaybackMode.loop

    def __len__(self):
        return len(self._songs)

    def __getitem__(self, index):
        """overload [] operator"""
        return self._songs[index]

    def mark_as_bad(self, song):
        if song in self._songs and song not in self._bad_songs:
            self._bad_songs.append(song)

    def is_bad(self, song):
        return song in self._bad_songs

    def _add(self, song):
        if song in self._songs:
            return
        self._songs.append(song)
        length = len(self._songs)
        self.songs_added.emit(length-1, 1)

    def set_models(self, models, next_=False, fm=False):
        """
        .. versionadded: v3.7.13
        """
        self.clear()
        self.batch_add(models)
        if fm is False:
            self.mode = PlaylistMode.normal
        else:
            self.mode = PlaylistMode.fm
        if next_ is True:
            self.next()

    def batch_add(self, models):
        """
        .. versionadded: v3.7.13
        """
        start_index = len(self._songs)
        for model in models:
            self._songs.append(model)
        end_index = len(self._songs)
        self.songs_added.emit(start_index, end_index - start_index)

    def add(self, song):
        """add song to playlist

        Theoretically, when playlist is in FM mode, we should not
        change songs list manually(without ``fm_add`` method). However,
        when it happens, we exit FM mode.
        """
        if self._mode is PlaylistMode.fm:
            self.mode = PlaylistMode.normal
        self._add(song)

    def fm_add(self, song):
        self._add(song)

    def insert(self, song):
        """Insert song after current song

        When current song is none, the song is appended.
        """
        if self._mode is PlaylistMode.fm:
            self.mode = PlaylistMode.normal
        if song in self._songs:
            return
        if self._current_song is None:
            self._add(song)
        else:
            index = self._songs.index(self._current_song)
            self._songs.insert(index + 1, song)
            self.songs_added.emit(index + 1, 1)

    def remove(self, song):
        """Remove song from playlist. O(n)

        If song is current song, remove the song and play next. Otherwise,
        just remove it.
        """
        try:
            index = self._songs.index(song)
        except ValueError:
            logger.debug('Remove failed: {} not in playlist'.format(song))
        else:
            if self._current_song is None:
                self._songs.remove(song)
            elif song == self._current_song:
                next_song = self.next_song
                # 随机模式下或者歌单只剩一首歌曲，下一首可能和当前歌曲相同
                if next_song == self.current_song:
                    self.current_song = None
                    self._songs.remove(song)
                    self.current_song = self.next_song
                else:
                    next_song = self.next_song
                    self._songs.remove(song)
                    self.current_song = next_song
            else:
                self._songs.remove(song)
            self.songs_removed.emit(index, 1)
            logger.debug('Remove {} from player playlist'.format(song))
        if song in self._bad_songs:
            self._bad_songs.remove(song)

    def init_from(self, songs):
        warnings.warn(
            'use set_models instead, this will be removed in v3.8',
            DeprecationWarning
        )
        self.set_models(songs, fm=False)

    def clear(self):
        """remove all songs from playlists"""
        if self.current_song is not None:
            self.current_song = None
        length = len(self._songs)
        self._songs.clear()
        if length > 0:
            self.songs_removed.emit(0, length)
        self._bad_songs.clear()

    def list(self):
        """Get all songs in playlists"""
        return self._songs

    @property
    def playback_mode(self):
        return self._playback_mode

    @playback_mode.setter
    def playback_mode(self, playback_mode):
        if self._mode is PlaylistMode.fm:
            if playback_mode is not PlaybackMode.sequential:
                logger.warning("can't set playback mode to others in fm mode")
                return
        self._playback_mode = playback_mode
        self.playback_mode_changed.emit(self.playback_mode)

    def _get_good_song(self, base=0, random_=False, direction=1, loop=True):
        """从播放列表中获取一首可以播放的歌曲

        :param base: base index
        :param random: random strategy or not
        :param direction: forward if > 0 else backward
        :param loop: regard the song list as a loop

        >>> from unittest import mock
        >>> pl = Playlist(mock.Mock(), [1, 2, 3])
        >>> pl._get_good_song()
        1
        >>> pl._get_good_song(base=1)
        2
        >>> pl._bad_songs = [2]
        >>> pl._get_good_song(base=1, direction=-1)
        1
        >>> pl._get_good_song(base=1)
        3
        >>> pl._bad_songs = [1, 2, 3]
        >>> pl._get_good_song()
        """
        if not self._songs or len(self._songs) <= len(self._bad_songs):
            logger.debug('No good song in playlist.')
            return None

        good_songs = []
        if direction > 0:
            if loop is True:
                song_list = self._songs[base:] + self._songs[0:base]
            else:
                song_list = self._songs[base:]
        else:
            if loop is True:
                song_list = self._songs[base::-1] + self._songs[:base:-1]
            else:
                song_list = self._songs[base::-1]
        for song in song_list:
            if song not in self._bad_songs:
                good_songs.append(song)
        if not good_songs:
            return None
        if random_:
            return random.choice(good_songs)
        else:
            return good_songs[0]

    @property
    def next_song(self):
        """next song for player, calculated based on playback_mode"""
        # 如果没有正在播放的歌曲，找列表里面第一首能播放的
        if self.current_song is None:
            return self._get_good_song()

        if self.playback_mode == PlaybackMode.random:
            next_song = self._get_good_song(random_=True)
        else:
            current_index = self._songs.index(self.current_song)
            if current_index == len(self._songs) - 1:
                if self.playback_mode in (PlaybackMode.loop, PlaybackMode.one_loop):
                    next_song = self._get_good_song()
                elif self.playback_mode == PlaybackMode.sequential:
                    next_song = None
            else:
                next_song = self._get_good_song(base=current_index+1, loop=False)
        return next_song

    @property
    def previous_song(self):
        """previous song for player to play

        NOTE: not the last played song
        """
        if self.current_song is None:
            return self._get_good_song(base=-1, direction=-1)

        if self.playback_mode == PlaybackMode.random:
            previous_song = self._get_good_song(direction=-1, random_=True)
        else:
            current_index = self._songs.index(self.current_song)
            previous_song = self._get_good_song(base=current_index - 1, direction=-1)
        return previous_song

    async def a_next(self):
        self.next()

    def next(self) -> Optional[asyncio.Task]:
        if self.next_song is None:
            self.eof_reached.emit()
            return None
        else:
            return self.set_current_song(self.next_song)

    def _on_media_finished(self):
        # Play next model when current media is finished.
        if self.playback_mode == PlaybackMode.one_loop:
            return self.set_current_song(self.current_song)
        else:
            self.next()

    def _on_song_changed(self, song):
        self._app.task_mgr.run_afn_preemptive(self._fetch_current_song_mv, song)

    async def _fetch_current_song_mv(self, song):
        if song is None:
            self._current_song_mv = None
        else:
            try:
                mv = await run_fn(self._app.library.song_get_mv, song)
            except ProviderIOError:
                logger.exception('get song mv info failed')
                self._current_song_mv = None
            else:
                self._current_song_mv = mv
        self.song_mv_changed.emit(song, self._current_song_mv)

    def previous(self) -> Optional[asyncio.Task]:
        """return to the previous song in playlist"""
        return self.set_current_song(self.previous_song)

    @property
    def current_song(self) -> Optional[SongModel]:
        """Current song

        return None if there is no current song
        """
        return self._current_song

    @current_song.setter
    def current_song(self, song: Optional[SongModel]):
        self.set_current_song(song)

    @property
    def current_song_mv(self) -> Optional[VideoModel]:
        return self._current_song_mv

    def set_current_song(self, song) -> Optional[asyncio.Task]:
        """设置当前歌曲，将歌曲加入到播放列表，并发出 song_changed 信号

        .. note::

            该方法理论上只应该被 Player 对象调用。

        if song has not valid media, we find a replacement in other providers

        .. versionadded:: 3.7.11
           The method is added to replace current_song.setter.
        """
        if song is None:
            self.pure_set_current_song(None, None, None)
            return None

        if self.mode is PlaylistMode.fm and song not in self._songs:
            self.mode = PlaylistMode.normal

        # FIXME(cosven): `current_song.setter` depends on app.task_mgr and app.library,
        # which make it hard to test.
        return self._t_scm.bind_coro(self.a_set_current_song(song))

    async def a_set_current_song(self, song):
        """Set the `song` as the current song.

        If the song is bad, then this will try to use a standby in Playlist.normal mode.
        """
        target_song = song  # The song to be set.
        media = None        # The corresponding media to be set.
        try:
            media = await self._prepare_media(song)
        except MediaNotFound as e:
            if e.reason is MediaNotFound.Reason.check_children:
                await self.a_set_current_song_children(song)
                return
        except ProviderIOError as e:
            # FIXME: This may cause infinite loop when the prepare media always fails
            logger.error(f'prepare media failed: {e}, try next song')
            run_afn(self.a_next)
            return
        except Exception as e:  # noqa
            # When the exception is unknown, we mark the song as bad.
            self._app.show_msg(f'获取歌曲链接失败: {e}')
            logger.exception('prepare media failed due to unknown error')
        else:
            assert media, "media must not be empty"

        # The song has no media, try to find and use standby unless it is in fm mode.
        if media is None:
            if self._app.config.ENABLE_MV_AS_STANDBY:
                self._app.show_msg('尝试获取音乐视频的播放资源...')
                media = await self._prepare_mv_media(song)

            if media:
                self._app.show_msg('使用音乐视频作为其播放资源 ✅')
            else:
                # if mode is fm mode, do not find standby song, just skip the song.
                if self.mode is PlaylistMode.fm:
                    self.mark_as_bad(song)
                    run_afn(self.a_next)
                    return

                logger.info(f"no media found for {song}, mark it as bad")
                self.mark_as_bad(song)
                target_song, media = await self.find_and_use_standby(song)

        metadata = await self._prepare_metadata_for_song(target_song)
        self.pure_set_current_song(target_song, media, metadata)

    async def a_set_current_song_children(self, song):
        # TODO: maybe we can just add children to playlist?
        self._app.show_msg(f'{song} 的播放资源在孩子节点上，将孩子节点添加到播放列表')
        self.mark_as_bad(song)
        logger.info(f'{song} has children, replace the current playlist')
        song = await run_fn(self._app.library.song_upgrade, song)
        if song.children:
            self.batch_add(song.children)
            await self.a_set_current_song(song.children[0])
        else:
            run_afn(self.a_next)
        return

    async def find_and_use_standby(self, song):
        self._app.show_msg(f'{song} 无可用的播放资源, 尝试寻找备用歌曲...')
        logger.info(f'try to find standby from other providers for {song}')
        standby_candidates = await self._app.library.a_list_song_standby_v2(
            song,
            self.audio_select_policy
        )
        if standby_candidates:
            standby, media = standby_candidates[0]
            logger.info(f'song standby was found in {standby.source} ✅')
            self._app.show_msg(f'在 {standby.source} 平台找到 {song} 的备用歌曲 ✅')
            # Insert the standby song after the song
            if song in self._songs and standby not in self._songs:
                index = self._songs.index(song)
                self._songs.insert(index + 1, standby)
                self.songs_added.emit(index + 1, 1)
            return standby, media

        logger.info(f'{song} song standby not found')
        self._app.show_msg(f'未找到 {song} 的备用歌曲')
        return song, None

    def pure_set_current_song(self, song, media, metadata=None):
        if song is None:
            self._current_song = None
        else:
            # add it to playlist if song not in playlist
            if song in self._songs:
                self._current_song = song
            else:
                self.insert(song)
                self._current_song = song
        self.song_changed.emit(song)
        self.song_changed_v2.emit(song, media)

        if song is not None:
            if media is None:
                run_afn(self.a_next)
            else:
                # Note that the value of model v1 {}_display may be None.
                kwargs = {}
                if not self._app.has_gui:
                    kwargs['video'] = False
                # TODO: set artwork field
                self._app.player.play(media, metadata=metadata, **kwargs)
        else:
            self._app.player.stop()

    async def _prepare_metadata_for_song(self, song):
        metadata = Metadata({
            MetadataFields.uri: reverse(song),
            MetadataFields.source: song.source,
            MetadataFields.title: song.title_display or '',
            # The song.artists_name should return a list of strings
            MetadataFields.artists: [song.artists_name_display or ''],
            MetadataFields.album: song.album_name_display or '',
        })
        try:
            song = await aio.run_fn(self._app.library.song_upgrade, song)
        except ResourceNotFound:
            return metadata
        except:  # noqa
            logger.exception(f"fetching song's meta failed, song:'{song.title_display}'")
            return metadata

        artwork = ''
        released = ''
        if song.album is not None:
            try:
                album = await aio.run_fn(self._app.library.album_upgrade, song.album)
            except ResourceNotFound:
                pass
            except:  # noqa
                logger.warning(
                    f"fetching song's album meta failed, song:{song.title_display}")
            else:
                artwork = album.cover
                released = album.released
                # For model v1, the cover can be a Media object.
                # For example, in fuo_local plugin, the album.cover is a Media
                # object with url set to fuo://local/songs/{identifier}/data/cover.
                if isinstance(artwork, Media):
                    artwork = artwork.url

        # Try to use album meta first.
        if artwork and released:
            metadata[MetadataFields.artwork] = artwork
            metadata[MetadataFields.released] = released
        else:
            metadata[MetadataFields.artwork] = song.pic_url or artwork
            metadata[MetadataFields.released] = song.date or released
        return metadata

    async def _prepare_metadata_for_video(self, video):
        metadata = Metadata({
            # The value of model v1 title_display may be None.
            MetadataFields.title: video.title_display or '',
            MetadataFields.source: video.source,
            MetadataFields.uri: reverse(video),
        })
        try:
            video = await aio.run_fn(self._app.library.video_upgrade, video)
        except ModelNotFound as e:
            logger.warning(f"can't get cover of video due to {str(e)}")
        else:
            metadata[MetadataFields.artwork] = video.cover
        return metadata

    async def _prepare_media(self, song):
        task_spec = self._app.task_mgr.get_or_create('prepare-media')
        task_spec.disable_default_cb()
        if self.watch_mode is True:
            mv_media = await task_spec.bind_coro(self._prepare_mv_media(song))
            if mv_media:
                return mv_media
            self._app.show_msg('未找到可用的歌曲视频资源')
        return await task_spec.bind_blocking_io(
            self._app.library.song_prepare_media, song, self.audio_select_policy)

    async def _prepare_mv_media(self, song) -> Optional[Media]:
        try:
            mv_media = await run_fn(
                self._app.library.song_prepare_mv_media,
                song,
                self._app.config.VIDEO_SELECT_POLICY)
        except MediaNotFound:
            mv_media = None
        except Exception as e:  # noqa
            mv_media = None
            logger.exception(f'fail to get {song} mv: {e}')
        return mv_media

    def set_current_model(self, model):
        """
        .. versionadded: 3.7.13
        """
        if model is None:
            self._app.player.stop()
            return
        if ModelType(model.meta.model_type) is ModelType.song:
            return self.set_current_song(model)
        return self._t_scm.bind_coro(self.a_set_current_model(model))

    async def a_set_current_model(self, model):
        """
        TODO: handle when model is a song

        .. versionadded: 3.7.13
        """
        assert ModelType(model.meta.model_type) is ModelType.video, \
            "{model.meta.model_type} is not supported, expecting a video model, "

        video = model
        try:
            media = await aio.run_fn(
                self._app.library.video_prepare_media,
                video,
                self._app.config.VIDEO_SELECT_POLICY
            )
        except MediaNotFound:
            self._app.show_msg('没有可用的播放链接')
        else:
            metadata = await self._prepare_metadata_for_video(video)
            kwargs = {}
            if not self._app.has_gui:
                kwargs['video'] = False
            self._app.player.play(media, metadata=metadata, **kwargs)

    """
    Sync methods.

    Currently, playlist has both async and sync methods to keep backward
    compatibility. Sync methods will be replaced by async methods in the end.
    """
    def play_model(self, model):
        """Set current model and play it

        .. versionadded: 3.7.14
        """
        # Stop the player so that user know the action is working.
        self._app.player.stop()
        self.play_model_handling.emit()
        task = self.set_current_model(model)
        if task is not None:
            def cb(future):
                try:
                    future.result()
                except:  # noqa
                    logger.exception('play model failed')
                else:
                    self._app.player.resume()
                    logger.info(f'play a model ({model}) succeed')
            task.add_done_callback(cb)



from __future__ import annotations
import importlib
import logging
import os
import sys
from typing import TYPE_CHECKING

from feeluown.config import Config
from feeluown.utils.dispatch import Signal
from .consts import USER_PLUGINS_DIR

if TYPE_CHECKING:
    from feeluown.app import App


__all__ = (
    'plugins_mgr',
    'Plugin',
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class InvalidPluginError(Exception):
    pass


class Plugin:
    """
    A plugin can be a Python module or package which implements
    `enable(app)` and `disable(app)` function. It can also implements
    `init_config(config)` and initialize its configurations.
    """
    def __init__(self, module, alias='', version='', desc='',
                 author='', homepage='', dist_name=''):
        """插件对象

        :param alias: 插件名
        :param version: 插件版本
        :param module: 插件模块对象，它有 enable 和 disable 方法
        :param desc: 插件描述
        :param author: 插件作者
        :param homepage: 插件主页
        :param dist_name: 插件发行版名字
        """
        # pylint: disable=too-many-arguments

        self.alias = alias
        # FIXME(cosven): use entry point name as plugin name, instead of the module name.
        self.name = module.__name__.split('.')[-1]
        self._module = module
        self.version = version
        self.desc = desc
        self.author = author
        self.homepage = homepage
        self.dist_name = dist_name
        self.is_enabled = False

    @classmethod
    def create(cls, module):
        """Plugin 工厂函数

        :param module:
        :return:
        """
        try:
            # alias, desc, version 为必需字段
            alias = module.__alias__
            desc = module.__desc__
            _ = module.__version__  # noqa

            author = getattr(module, '__author__', '')
            homepage = getattr(module, '__homepage__', '')
            dist_name = getattr(module, '__dist_name__', '')
        except AttributeError as e:
            raise InvalidPluginError(str(e)) from None
        else:
            return Plugin(module,
                          alias=alias,
                          desc=desc,
                          author=author,
                          homepage=homepage,
                          dist_name=dist_name)

    def init_config(self, config: Config):
        """Call plugin.init_config function if possible

        :param config: app config instance.

        .. versionadded: 3.7.15
        """
        myconfig = Config()

        names = [self.name]
        # Currently, plugin name looks like fuo_xxx and xxx is the real name.
        # User maye want to define config like app.xxx.X=Y,
        # instead of app.fuo_xxx.X=Y.
        if self.name.startswith('fuo_'):
            names.append(self.name[4:])

        # Define a subconfig(namespace) for plugin so that plugin can
        # define its own configuration fields.
        for name in names:
            config.deffield(name,
                            type_=Config,
                            default=myconfig,
                            desc=f'Configurations for plugin {self.name}')

        try:
            fn = self._module.init_config
        except AttributeError:
            # The plugin does not define any config field.
            pass
        else:
            fn(myconfig)

    def enable(self, app):
        self._module.enable(app)
        self.is_enabled = True

    def disable(self, app):
        self._module.disable(app)
        self.is_enabled = False


class PluginsManager:
    def __init__(self):
        self._plugins = {}
        #: A plugin is about to enable.
        # The payload is the plugin object `(Plugin)`.
        # .. versionadded: 3.7.15
        self.about_to_enable = Signal()
        # scan_finished means all found plugins are enabled.
        # TODO: maybe rename scan_finished to plugins_enabled?
        self.scan_finished = Signal()

    def light_scan(self):
        """Scan plugins without enabling them."""
        logger.info('Light scan plugins.')
        self._scan_dirs()
        self._scan_entry_points()

    def init_plugins_config(self, config):
        """Try to init config for the plugin.

        Plugin can declare their configuration items.
        """
        for plugin in self._plugins.values():
            try:
                plugin.init_config(config)
            except Exception:  # noqa
                logger.exception(f'Init config for plugin:{plugin.name} failed')

    def enable_plugins(self, app: App):
        logger.info(f'Enable plugins that are scaned. total: {len(self._plugins)} ')
        for plugin in self._plugins.values():
            # Try to enbale the plugin.
            self.about_to_enable.emit(plugin)
            try:
                plugin.enable(app)
            except Exception:  # noqa
                logger.exception(f'Enable plugin:{plugin.name} failed')
        self.scan_finished.emit(list(self._plugins.values()))

    def load_plugin_from_module(self, module):
        """Load module and try to load the plugin"""
        logger.info('Try to load plugin from module: %s', module.__name__)

        # Try to create a new plugin.
        try:
            plugin = Plugin.create(module)
        except InvalidPluginError:
            return
        self._plugins[plugin.name] = plugin

    def _scan_dirs(self):
        """扫描插件目录中的插件"""
        if not os.path.exists(USER_PLUGINS_DIR):
            return

        module_name_list = []
        for fname in os.listdir(USER_PLUGINS_DIR):
            if os.path.isdir(os.path.join(USER_PLUGINS_DIR, fname)):
                module_name_list.append(fname)
            else:
                if fname.endswith('.py'):
                    module_name_list.append(fname[:-3])
        sys.path.append(USER_PLUGINS_DIR)
        for module_name in module_name_list:
            try:
                module = importlib.import_module(module_name)
            except Exception:  # noqa
                logger.exception('Failed to import module %s', module_name)
            else:
                self.load_plugin_from_module(module)

    def _scan_entry_points(self):
        """扫描通过 setuptools 机制注册的插件

        https://packaging.python.org/guides/creating-and-discovering-plugins/
        """
        try:
            import importlib.metadata  # pylint: disable=redefined-outer-name
            try:
                entry_points = importlib.metadata.entry_points() \
                    .select(group='fuo.plugins_v1')
            except AttributeError:  # old version does not has `select` method
                entry_points = importlib.metadata.entry_points() \
                    .get('fuo.plugins_v1', [])
        except ImportError:
            import pkg_resources
            entry_points = pkg_resources.iter_entry_points('fuo.plugins_v1')
        for entry_point in entry_points:
            try:
                module = entry_point.load()
            except Exception:  # noqa
                logger.exception('Failed to load module %s', entry_point.name)
            else:
                self.load_plugin_from_module(module)


plugins_mgr = PluginsManager()



from functools import partial

from .gateway import Gateway


class LiveLyricPublisher:
    topic = 'live_lyric'

    def __init__(self, gateway: Gateway):
        self.gateway = gateway
        gateway.add_topic(self.topic)

    def publish(self, sentence):
        self.gateway.publish(sentence + '\n', self.topic)


class SignalPublisher:
    def __init__(self, gateway: Gateway):
        self.gateway = gateway

    def on_emitted(self, name):
        return partial(self.publish, name)

    def publish(self, name, *args):
        self.gateway.publish(list(args), name, need_serialize=True)



import base64
import json
import logging
import os
import re
from typing import Optional

from pydantic import ValidationError
from mutagen import MutagenError
from mutagen.mp3 import EasyMP3
from mutagen.easymp4 import EasyMP4
from mutagen.flac import FLAC
from mutagen.apev2 import APEv2File

from feeluown.serializers import serialize
from feeluown.utils.utils import elfhash, log_exectime
from feeluown.utils.lang import can_convert_chinese, convert_chinese
from feeluown.library import SongModel, AlbumModel, ArtistModel, AlbumType
from feeluown.library import BriefAlbumModel, BriefArtistModel, BriefSongModel
from feeluown.library import reverse

from .schemas import EasyMP3Model, APEModel, FLACModel
from .schemas import DEFAULT_ALBUM_NAME


logger = logging.getLogger(__name__)
SOURCE = 'local'


def to_brief_song(song):
    return BriefSongModel(
        identifier=song.identifier,
        source=song.source,
        title=song.title,
        artists_name=song.artists_name,
        album_name=song.album_name,
        duration_ms=song.duration_ms,
    )


def to_brief_album(album):
    return BriefAlbumModel(
        identifier=album.identifier,
        source=album.source,
        name=album.name,
        artists_name=album.artists_name,
    )


def to_brief_artist(artist):
    return BriefArtistModel(
        identifier=artist.identifier,
        source=artist.source,
        name=artist.name,
    )


def gen_id(s):
    return str(elfhash(base64.b64encode(bytes(s, 'utf-8'))))


def gen_cover_url(song):
    return reverse(song) + '/cover/data'


def create_artist(identifier, name):
    return ArtistModel(identifier=identifier,
                       source=SOURCE,
                       name=name,
                       pic_url='',
                       aliases=[],
                       hot_songs=[],
                       description='')


def create_album(identifier, name, cover):
    album = AlbumModel(identifier=identifier,
                       source=SOURCE,
                       name=name,
                       cover=cover or '',
                       songs=[],
                       artists=[],
                       description='')
    # guess album type by its name
    #
    # album name which contains following string are `Single`
    #   1. ' - Single'  6+3=9
    #   2. '(single)'   6+2=8
    #   3. '（single）'  6+2=8
    if 'single' in name[-9:].lower():
        album.type_ = AlbumType.single
    if 'ep' in name[-5:].lower():
        album.type_ = AlbumType.ep
    return album


def read_audio_metadata(fpath, can_convert_chinese=False, lang='auto') -> Optional[dict]:
    """Read metadata, like id3tag, from audio file.

    Return a dict if succeed. Return None when failed, no exception is raised.
    The dict schema is `schemas.Common`.
    """
    try:
        if fpath.endswith('mp3') or fpath.endswith('ogg') or fpath.endswith('wma'):
            metadata = EasyMP3(fpath)
        elif fpath.endswith('m4a') or fpath.endswith('m4v') or fpath.endswith('mp4'):
            metadata = EasyMP4(fpath)
        elif fpath.endswith('flac'):
            metadata = FLAC(fpath)
        elif fpath.endswith('ape'):
            metadata = APEv2File(fpath)
        elif fpath.endswith('wav'):
            metadata = None
        else:  # This branch is actually impossible to reach.
            metadata = None
    except MutagenError as e:
        logger.warning(
            'Mutagen parse metadata failed, ignore.\n'
            'file: {}, exception: {}'.format(fpath, str(e)))
        return

    # For example::
    #
    #   {
    #    'album': ['Beyond Live 1991 生命接触演唱会'],
    #    'title': ['再见理想'],
    #    'artist': ['BEYOND'],
    #    'genre': ['Blues']
    #   }
    metadata_dict = dict(metadata or {})
    metadata_dict = {k: v[0] for k, v in metadata_dict.items()}

    # Get title and artists name from filename when metadata is empty.
    if metadata is None:
        filename = os.path.split(fpath)[-1].split('.')[0]
        parts = filename.split(' - ', 1)
        if len(parts) == 2:
            title, artists_name = parts
            metadata_dict['artists_name'] = artists_name
        else:
            title = parts[0]
        metadata_dict['title'] = title
        metadata_dict['duration'] = 0
    else:
        # milesecond
        metadata_dict['duration'] = int(metadata.info.length * 1000)

    # Convert simplified to traditional, or reverse.
    if can_convert_chinese:
        metadata_dict = {k: convert_chinese(v, lang) for k, v in metadata_dict.items()}

    try:
        if fpath.endswith('flac'):
            data = FLACModel(**metadata_dict).dict()
            data['track'] = f"{data['track_number']}/{data['track_total']}"
            data['disc'] = f"{data['disc_number']}/{data['disc_total']}"
        elif fpath.endswith('ape'):
            data = APEModel(**metadata_dict).dict()
        else:
            data = EasyMP3Model(**metadata_dict).dict()
    except ValidationError:
        logger.exception(f'parse audio file metadata ({fpath}) failed')
        return
    return data


def gen_artist_name_list(artists_name, splitter, splitter_ignorance):
    # For example::
    # artists_name: 'Years & Years & Jess Glynne'
    # splitter: [',', '&']
    # splitter_ignorance: ['Years & Years']
    # return: ['Years & Years', 'Jess Glynne']
    if splitter_ignorance is None:
        splitter_ignorance = []
    if splitter_ignorance:
        artists_name = re.split(r'({})'.format('|'.join(splitter_ignorance)),
                                artists_name)
    else:
        artists_name = [artists_name]
    artist_name_list = []
    for artist_name in artists_name:
        if artist_name in splitter_ignorance:
            artist_name_list.append(artist_name)
        else:
            artist_name_list.extend(re.split(r'{}'.format('|'.join(splitter)),
                                             artist_name))
    return [artist_name.strip()
            for artist_name in artist_name_list if artist_name.strip()]


def add_song(fpath, g_songs, g_artists, g_albums, g_file_song, g_album_contributors,
             can_convert_chinese=False, lang='auto',
             delimiter='', expand_artist_songs=False,
             artist_splitter=[',', '&'], artist_splitter_ignorance=None,
             split_album_artist_name=False):
    """
    parse music file metadata with Easymp3 and return a song
    model.
    """
    data = read_audio_metadata(fpath, can_convert_chinese, lang)
    if data is None:
        return

    # NOTE: use {title}-{artists_name}-{album_name} as song identifier
    title = data['title']
    album_name = data['album_name']
    artist_name_list = gen_artist_name_list(
        data['artists_name'], artist_splitter, artist_splitter_ignorance)
    artists_name = ','.join(artist_name_list)
    duration = data['duration']
    album_artist_name = data['album_artist_name']

    # 生成 song model
    # 用来生成 id 的字符串应该尽量减少无用信息，这样或许能减少 id 冲突概率
    # 加入分隔符'-'在一定概率上更能确保不发生哈希值重复
    song_id_str = delimiter.join([title, artists_name, album_name, str(int(duration))])
    song_id = gen_id(song_id_str)
    if song_id not in g_songs:
        song = SongModel(identifier=song_id,
                         source=SOURCE,
                         artists=[],
                         title=title,
                         duration=duration,
                         genre=data['genre'],
                         date=data['date'],
                         disc=data['disc'],
                         track=data['track'])
        g_file_song[fpath] = song_id
        g_songs[song_id] = song
    else:
        song = g_songs[song_id]
        logger.warning('Duplicate song: %s', fpath)
        return

    # 生成 album artist model
    if split_album_artist_name:
        album_artist_name_list = gen_artist_name_list(
            album_artist_name, artist_splitter, artist_splitter_ignorance)
    else:
        album_artist_name_list = [album_artist_name]

    album_artist_id_list, album_artist_list = [], []
    for artist_name in album_artist_name_list:
        artist_id = gen_id(artist_name)
        if artist_id not in g_artists:
            album_artist = create_artist(artist_id, artist_name)
            g_artists[artist_id] = album_artist
        else:
            album_artist = g_artists[artist_id]
        album_artist_id_list.append(artist_id)
        album_artist_list.append(album_artist)

    # 生成 album model
    album_id_str = delimiter.join([album_name, album_artist_name])
    album_id = gen_id(album_id_str)
    cover = gen_cover_url(song)
    if album_id not in g_albums:
        album = create_album(album_id, album_name, cover)
        g_albums[album_id] = album
        g_album_contributors[album_id] = []
    else:
        album = g_albums[album_id]

    # 处理专辑的歌手信息和歌曲信息，专辑歌手的专辑列表信息
    for album_artist in album_artist_list:
        for artist in album.artists:
            if album_artist.identifier == artist.identifier:
                break
        else:
            album.artists.append(to_brief_artist(album_artist))

    if song not in album.songs:
        album.songs.append(song)

    # 处理歌曲的歌手和专辑信息，以及歌手的歌曲列表和参与作品
    song.album = to_brief_album(album)
    for artist_name in artist_name_list:
        artist_id = gen_id(artist_name)
        if artist_id in g_artists:
            artist = g_artists[artist_id]
        else:
            artist = create_artist(identifier=artist_id, name=artist_name)
            g_artists[artist_id] = artist
        if artist not in song.artists:
            song.artists.append(to_brief_artist(artist))
        if song not in artist.hot_songs:
            artist.hot_songs.append(song)

        # 处理歌曲歌手的参与作品信息(不与前面的重复)
        if artist_id not in album_artist_id_list \
                and artist_id not in g_album_contributors[album_id]:
            g_album_contributors[album_id].append(artist_id)

    # 处理专辑歌手的歌曲信息: 有些作词人出合辑很少出现在歌曲歌手里(可选)
    for album_artist_ in album_artist_list:
        if expand_artist_songs and song not in album_artist_.hot_songs:
            album_artist_.hot_songs.append(song)


def scan_directory(directory, exts, depth=2):
    if depth < 0:
        return []

    media_files = []
    if not os.path.exists(directory):
        return []
    for path in os.listdir(directory):
        path = os.path.join(directory, path)
        if os.path.isdir(path):
            files = scan_directory(path, exts, depth - 1)
            media_files.extend(files)
        elif os.path.isfile(path):
            if path.split('.')[-1] in exts:
                media_files.append(path)
    return media_files


def sort_album_func(album):
    if album.songs:
        return album.songs[0].date is not None, album.songs[0].date
    return False, '0'


class DB:
    """
    DB manages a fileset and their corresponding models

    the data structure in db file::

        {
          "files": [],
          "songs": []
        }
    """
    def __init__(self, fpath):
        """
        :param filepath: database file path
        """

        self._dirty = False  # whether changes are flushed
        self._fileset = set()  # media files
        self._fpath = fpath

        self._file_song = {}           # {fpath: song_id}
        self._song_file = {}           # {song_id: fpath}
        self._songs = {}               # {song_id: song}
        self._albums = {}              # {album_id: album)
        self._artists = {}             # {artist_id: artist)
        self._album_contributors = {}  # {album_id: [artist_id]}

    def flush(self):
        """flush the changes into db file"""
        with open(self._fpath, 'w', encoding='utf-8') as f:
            json.dump({
                'version': '1.0',
                'files': list(self._fileset),
                'songs': serialize('json', self._songs),
                'artists': serialize('json', self._artists),
                'albums': serialize('json', self._albums),
            }, f)

    def list_models(self):
        """list all models in database"""
        pass

    def add(self, fpath):
        """add media file to database"""

    def remove(self, fpath):
        """add media file to database"""

    ############################

    def list_songs(self):
        return list(self._songs.values())

    def list_albums(self):
        return list(self._albums.values())

    def list_albums_by_contributor(self, artist_id):
        albums = []
        for album_id, artists in self._album_contributors.items():
            if artist_id in artists:
                album = self.get_album(album_id)
                albums.append(album)
        return albums

    def list_artists(self):
        return list(self._artists.values())

    def get_song(self, identifier):
        return self._songs.get(identifier)

    def get_album(self, identifier):
        return self._albums.get(identifier)

    def get_artist(self, identifier):
        return self._artists.get(identifier)

    def get_song_fpath(self, song_id) -> Optional[str]:
        return self._song_file.get(song_id)

    @log_exectime
    def scan(self, config, paths, depth, exts):
        """scan media files in all paths
        """
        media_files = []
        logger.info('start scanning...')
        for directory in paths:
            logger.debug('正在扫描目录(%s)...', directory)
            media_files.extend(scan_directory(directory, exts, depth))
        logger.info(f'scanning finished, {len(media_files)} files in total')

        is_cn_convert_enabled = can_convert_chinese()
        for fpath in media_files:
            add_song(fpath, self._songs, self._artists,
                     self._albums, self._file_song, self._album_contributors,
                     is_cn_convert_enabled, config.CORE_LANGUAGE,
                     config.IDENTIFIER_DELIMITER, config.EXPAND_ARTIST_SONGS,
                     config.ARTIST_SPLITTER, config.ARTIST_SPLITTER_IGNORANCE,
                     config.SPLIT_ALBUM_ARTIST_NAME)
        logger.info('录入本地音乐库完毕')

    def after_scan(self):
        # Sort the songs in a album.
        for album in self._albums.values():
            try:
                album.songs.sort(key=lambda x: (int(x.disc.split('/')[0]),
                                                int(x.track.split('/')[0])))
                if album.name != DEFAULT_ALBUM_NAME:
                    cover = gen_cover_url(album.songs[0])
                    album.cover = cover
            except:  # noqa
                logger.exception('Sort album songs failed.')

        # Select a pic_url for the artist
        for artist in self._artists.values():
            albums = []
            for album in self._albums.values():
                for artist_ in album.artists:
                    if artist_.identifier == artist.identifier:
                        albums.append(album)
                        continue
            if albums:
                albums.sort(key=sort_album_func, reverse=True)
                if albums:
                    artist.pic_url = albums[0].cover

            if not artist.pic_url and artist.hot_songs:
                # sort the artist hot_songs.
                artist.hot_songs.sort(key=lambda x: x.title)
                # Use a song's cover as artist cover.
                # https://github.com/feeluown/feeluown-local/pull/3/files#r362126996
                songs_with_unknown_album = [song for song in artist.hot_songs
                                            if song.album_name == DEFAULT_ALBUM_NAME]
                for song in sorted(songs_with_unknown_album,
                                   key=lambda x: (x.date != '', x.date),
                                   reverse=True):
                    artist.pic_url = gen_cover_url(song)
                    break

        # Cache the {song_id:fpath} mapping.
        self._song_file = {v: k for k, v in self._file_song.items()}
